@@ -11,8 +11,8 @@ readParmayMatrix = function(filename, output = c("all", "sf", "dynrange")[1],
   
   # The header information contains:
   #  - remarks (empty in all files I have tested): bytes 1:512
-  #  - Count of the total number of packets in file (bytes 513:516)
-  #  - header string = "MDTC" (bytes 517:520, if not there -> file corrupt)
+  #  - header string = "MDTC" (bytes 513:516, if not there -> file corrupt)
+  #  - Count of the total number of packets in file (bytes 517:520)
   #  - range of the acc (bytes 521:522) and the gyro sensors (bytes 523:524)
   
   # Each packet contains the following information:
@@ -35,8 +35,112 @@ readParmayMatrix = function(filename, output = c("all", "sf", "dynrange")[1],
   #        - checksum is used to verify integrity of data in each packet 
   #        - resampling is used with resample function to make the frequency stable
   #  - Gaps in data are not expected
-  
   # -------------------------------------------------------------------------
+  # Helper function to resample matrix sensor signals onto a common time grid, 
+  # usually the accelerometer's reference timeline.
+  # data = matrix with the sensor data
+  # timestamps = numeric vector of timestamps corresponding to rows fo `data`
+  # grid = numeric vector of target timestamps (usually, accelerometer resampled timestamps)
+  # interpolationType = 
+  resample_to_grid = function(data, timestamps, grid, interpolationType = 1) {
+    
+    ## ---- 0. Coerce shapes early ----------------------------------------------
+    if (is.matrix(timestamps) || is.data.frame(timestamps)) {
+      timestamps = timestamps[, 1]
+    }
+    timestamps = as.vector(timestamps)
+    data = as.matrix(data)
+    
+    common_len = min(NROW(data), length(timestamps))
+    if (common_len == 0L) {
+      out = matrix(NA_real_, nrow = length(grid), ncol = NCOL(data))
+      colnames(out) = colnames(data)
+      return(out)
+    }
+    data = data[seq_len(common_len), , drop = FALSE]
+    timestamps = timestamps[seq_len(common_len)]
+    
+    ## ---- 1. Normalize input types ------------------------------------------
+    # Convert POSIXct timestamps to numeric (seconds since epoch) if necessary
+    is_posix = inherits(timestamps, "POSIXct")
+    tz = if (is_posix) attr(timestamps, "tzone") else NULL
+    ts_num = if (is_posix) as.numeric(timestamps) else as.numeric(timestamps)
+    grid_num = if (inherits(grid, "POSIXct")) as.numeric(grid) else as.numeric(grid)
+    
+    data = as.matrix(data)
+    
+    ## ---- 2. Clean and sort input data --------------------------------------
+    # Remove rows with NA timestamps or NA values in any data column
+    ok = is.finite(ts_num) & rowSums(!is.finite(data)) == 0
+    ts_num = ts_num[ok]
+    data = data[ok, , drop = FALSE]
+    
+    # Ensure strictly increasing timestamps
+    ord = order(ts_num)
+    ts_num = ts_num[ord]
+    data = data[ord, , drop = FALSE]
+    
+    ## ---- 3. Remove duplicate timestamps ------------------------------------
+    # Duplicate timestamps cause division-by-zero errors in interpolation.
+    # Combine duplicates by taking the column-wise mean.
+    if (any(duplicated(ts_num))) {
+      split_ix = split(seq_along(ts_num), ts_num)
+      ts_unique = as.numeric(names(split_ix))
+      data_unique = do.call(rbind, lapply(split_ix, function(ix) {
+        colMeans(data[ix, , drop = FALSE])
+      }))
+      ord2 = order(ts_unique)
+      ts_num = ts_unique[ord2]
+      data = data_unique[ord2, , drop = FALSE]
+    }
+    
+    ## ---- 4. Ensure matching length and valid range -------------------------
+    # The C++ function assumes length(rawTime) == nrow(raw)
+    n_raw = min(length(ts_num), nrow(data))
+    if (n_raw < 2L) {
+      # Not enough data to interpolate -> return all NA
+      out = matrix(NA_real_, nrow = length(grid_num), ncol = ncol(data))
+      colnames(out) = colnames(data)
+      return(out)
+    }
+    
+    ts_num = ts_num[seq_len(n_raw)]
+    data = data[seq_len(n_raw), , drop = FALSE]
+    t0 = ts_num[1]
+    t1 = ts_num[n_raw]
+    
+    ## ---- 5. Clip the grid to the sensor s valid range ----------------------
+    # Only pass grid points within [t0, t1] to C++ for interpolation.
+    # The rest will be filled with NA later.
+    inside = (grid_num >= t0) & (grid_num <= t1)
+    grid_in = grid_num[inside]
+    
+    ## ---- 6. Interpolate within valid range ---------------------------------
+    res_in = matrix(NA_real_, nrow = sum(inside), ncol = ncol(data))
+    if (length(grid_in) > 0L) {
+      res_in = resample(
+        raw     = data,
+        rawTime = ts_num,
+        time    = grid_in,
+        stop    = nrow(data),
+        type    = interpolationType
+      )
+    }
+    
+    ## ---- 7. Pad output to full grid length ---------------------------------
+    # Build output with same number of rows as the target grid.
+    # Fill with NA outside sensor coverage, and insert resampled data inside.
+    out = matrix(NA_real_, nrow = length(grid_num), ncol = ncol(data))
+    if (length(grid_in) > 0L) {
+      out[inside, ] = res_in
+    }
+    
+    # Preserve column names for clarity in downstream use
+    colnames(out) = colnames(data)
+    out
+  }
+  # -------------------------------------------------------------------------
+  # Main code
   # 1 - read file and extract header information
 
   # open connection to read the binary file
@@ -57,11 +161,9 @@ readParmayMatrix = function(filename, output = c("all", "sf", "dynrange")[1],
   total_packets = readBin(tPackets_bytes, "integer", size = 4, endian = "little")
   
   # Validate start and end packets
-  lastchunk = FALSE
   if (is.null(end)) end = total_packets
   if (end >= total_packets) {
     end = total_packets
-    lastchunk = TRUE
   }
   if (start < 1) start = 1
   if (start > total_packets) return(NULL)
@@ -83,10 +185,26 @@ readParmayMatrix = function(filename, output = c("all", "sf", "dynrange")[1],
   packet_starti = find_matrix_packet_start(filename, packet_header_bytes)
   packet_endi = c(packet_starti[-1] - 1, file.info(filename)$size)
   
-  # validate packet headers (i.e., nr. of appearances of MDTCPACK in files)
-  if (length(packet_starti) != total_packets) {
-    stop(paste0(basename(filename), ": Probably corrupted file"))
-  }
+  # save number of packets declared and observed
+  n_declared = total_packets
+  n_observed_file = length(packet_starti)
+  
+  # Use all packets available in the file (even if empty packets have been added at the end)
+  n_use = n_observed_file
+  packet_starti = packet_starti[seq_len(n_use)]
+  packet_endi = packet_endi[seq_len(n_use)]
+  
+  # Apply chunking AFTER truncation
+  i_from = max(1, start)
+  i_to = n_use
+  idx = seq.int(i_from, i_to)
+  n_rows = length(idx)
+  
+  packet_starti = packet_starti[idx]
+  packet_endi   = packet_endi[idx]
+  
+  # lastchunk should reflect the truncated space
+  lastchunk = (i_to == n_use)
   
   # start time
   # starttime_indices = (packet_starti[1] + 12):(packet_starti[1] + 15)
@@ -99,10 +217,6 @@ readParmayMatrix = function(filename, output = c("all", "sf", "dynrange")[1],
     starttime_posix = as.POSIXct(starttime, origin = "1970-1-1", tz = configtz)
   }  
   
-  # Select chunk of interest
-  packet_starti = packet_starti[start:end]
-  packet_endi = packet_endi[start:end]
-  
   # -------------------------------------------------------------------------
   # 3 - Check packets integrity with CRC32 checksum
   
@@ -114,15 +228,15 @@ readParmayMatrix = function(filename, output = c("all", "sf", "dynrange")[1],
   crc32_stored = ifelse(crc32_stored_signed < 0, crc32_stored_signed + 2^32, crc32_stored_signed)
   
   # observed CRC32
-  crc32_observed <- vapply(seq_along(packet_starti), function(i) {
+  crc32_observed = vapply(seq_along(packet_starti), function(i) {
     seek(con, packet_starti[i] + 11, "start")  # Move to correct position
     
-    size <- packet_endi[i] - packet_starti[i] - 11  # Calculate the actual size
+    size = packet_endi[i] - packet_starti[i] - 11  # Calculate the actual size
     if (size <= 0) return(NA_real_)  # Handle edge cases where size is invalid
     
-    crc32_observed_raw <- readBin(con, what = "raw", n = size, endian = "little")
+    crc32_observed_raw = readBin(con, what = "raw", n = size, endian = "little")
     
-    crc32_hex <- digest::digest(crc32_observed_raw, algo = "crc32", serialize = FALSE)
+    crc32_hex = digest::digest(crc32_observed_raw, algo = "crc32", serialize = FALSE)
     as.numeric(paste0("0x", crc32_hex))
   }, numeric(1))
   
@@ -131,16 +245,19 @@ readParmayMatrix = function(filename, output = c("all", "sf", "dynrange")[1],
   
   # Initialize Quality-Check (QC) data frame
   QClog = data.frame(
-    checksum_pass = rep(TRUE, length(start:end)),
-    blockID = start:end,
-    start = integer(length(start:end)),
-    end = integer(length(start:end)),
-    blockLengthSeconds = numeric(length(start:end)),
-    frequency_set = numeric(length(start:end)),
-    frequency_observed = numeric(length(start:end)),
-    imputed = logical(length(start:end)),
-    gap_with_previous_block_secs = numeric(length(start:end)), 
-    start_time_adjustment_secs = numeric(length(start:end))
+    checksum_pass = rep(TRUE, n_rows),
+    blockID = idx,
+    start = integer(n_rows),
+    end = integer(n_rows),
+    blockLengthSeconds = numeric(n_rows),
+    frequency_set = numeric(n_rows),
+    frequency_observed = numeric(n_rows),
+    imputed = logical(n_rows),
+    gap_with_previous_block_secs = numeric(n_rows), 
+    start_time_adjustment_secs = numeric(n_rows),
+    declared_packets = rep(n_declared, n_rows),
+    observed_packets = rep(n_observed_file, n_rows),
+    acc_recorded = TRUE
   )
   
   # log corrupt packets
@@ -182,7 +299,6 @@ readParmayMatrix = function(filename, output = c("all", "sf", "dynrange")[1],
     seek(con, packet_starti[i] + 31, "start")  # Move to correct position
     readBin(con, what = "integer", size = 4, endian = "little")
   }, numeric(1))
-  
   # Process timestamps and calculate sampling frequency
   packets_dur_s = packets_t1 - packets_t0
   sf_acc_observed = acc_count / packets_dur_s
@@ -195,6 +311,20 @@ readParmayMatrix = function(filename, output = c("all", "sf", "dynrange")[1],
   QClog[, "frequency_set"] = sf
   QClog[, "frequency_observed"] = sf_acc_observed
   
+  # At the moment the function requires that accelerometer data has been
+  # recorded to work, as accelerometer timestamps are used as reference time
+  if (sum(acc_count) == 0) {
+    QClog[, "acc_recording"] = FALSE
+    return(list(
+      QClog = QClog,
+      data = matrix(nrow = 0, ncol = 3),
+      header = list(sf = sf,
+                    acc_dynrange = acc_dynrange,
+                    starttime = starttime_posix),
+      lastchunk = TRUE
+    ))
+  }
+  
   # build data structure lists
   # acc
   acc_starts = packet_starti + 36; acc_stops = acc_starts + 6*acc_count - 1
@@ -202,7 +332,12 @@ readParmayMatrix = function(filename, output = c("all", "sf", "dynrange")[1],
     acc_readings = lapply(seq_along(acc_starts), function(i) {
       seek(con, acc_starts[i] - 1, "start")
       readings = readBin(con, what = "integer", size = 2, n = 3 * acc_count[i], endian = "little")
-      matrix(readings * (acc_dynrange / 32767), ncol = 3, byrow = T)
+      # make sure number of readings is multiple of 3 (might not be in corrupted packets)
+      n_triplets = length(readings) %/% 3
+      readings_ok = 3 * n_triplets
+      readings = readings[1:readings_ok]
+      denominator = ifelse(readings >= 0, 32767, 32768)
+      matrix(readings * (acc_dynrange / denominator), ncol = 3, byrow = T)
     })
   }
   
@@ -214,7 +349,12 @@ readParmayMatrix = function(filename, output = c("all", "sf", "dynrange")[1],
       gyro_readings = lapply(seq_along(gyro_starts), function(i) {
         seek(con, gyro_starts[i] - 1, "start")  
         readings = readBin(con, what = "integer", size = 2, n = 3 * gyro_count[i], endian = "little")
-        matrix(readings * (gyro_range / 32767), ncol = 3, byrow = T)
+        # make sure number of readings is multiple of 3 (might not be in corrupted packets)
+        n_triplets = length(readings) %/% 3
+        readings_ok = 3 * n_triplets
+        readings = readings[1:readings_ok]
+        denominator = ifelse(readings >= 0, 32767, 32768)
+        matrix(readings * (gyro_range / denominator), ncol = 3, byrow = T)
       })
     }
     prev_stops = gyro_stops
@@ -226,6 +366,10 @@ readParmayMatrix = function(filename, output = c("all", "sf", "dynrange")[1],
       temp_readings = lapply(seq_along(temp_starts), function(i) {
         seek(con, temp_starts[i] - 1, "start")
         readings = readBin(con, what = "integer", size = 2, n = 2 * temp_count[i], endian = "little")
+        # make sure number of readings is multiple of 2 (might not be in corrupted packets)
+        n_triplets = length(readings) %/% 2
+        readings_ok = 2 * n_triplets
+        readings = readings[1:readings_ok]
         matrix(readings / 10, ncol = 2, byrow = T)
       })
     }
@@ -238,6 +382,10 @@ readParmayMatrix = function(filename, output = c("all", "sf", "dynrange")[1],
       heart_readings = lapply(seq_along(heart_starts), function(i) {
         seek(con, heart_starts[i] - 1, "start")
         readings = readBin(con, what = "integer", size = 2, n = 2 * heart_count[i], endian = "little")
+        # make sure number of readings is multiple of 3 (might not be in corrupted packets)
+        n_triplets = length(readings) %/% 2
+        readings_ok = 2 * n_triplets
+        readings = readings[1:readings_ok]
         matrix(readings, ncol = 2, byrow = T)
       })
     }
@@ -264,11 +412,45 @@ readParmayMatrix = function(filename, output = c("all", "sf", "dynrange")[1],
         x[corrupt_packets] = NA
         prev_index = ifelse(any(!is.na(x[1:(i - 1)])), 
                             max(x[1:(i - 1)], na.rm = T), NA)
-        prev_temp = temp_readings[[prev_index]][nrow(temp_readings[[prev_index]]), ]
+        if (!is.na(prev_index)) {
+          if (temp_count[prev_index] > 0) {
+            prev_temp = temp_readings[[prev_index]][nrow(temp_readings[[prev_index]]), ]
+          } else {
+            prev_temp = NA
+          }
+        } else {
+          # if the first packet is corrupted, then impute by first non-corrupted temperature observed
+          first_valid_temp = which(!1:length(temp_count) %in% corrupt_packets)[1]
+          if (is.na(first_valid_temp)) {
+            prev_temp = NA
+          } else {
+            prev_temp = temp_readings[[first_valid_temp]][1,]  
+          }
+        }
+        # next index
         next_index = ifelse(any(!is.na(x[(i + 1):length(x)])), 
                             min(x[(i + 1):length(x)], na.rm = T), NA)
-        next_temp = temp_readings[[next_index]][1, ]
-        mean_temp = (prev_temp + next_temp) / 2
+        if (!is.na(next_index)) {
+          if (temp_count[next_index] > 0) {
+            next_temp = temp_readings[[next_index]][1, ]
+          } else {
+            next_temp = NA
+          }
+        } else {
+          next_temp = NA
+        }
+        if (!is.na(prev_temp[1]) & !is.na(next_temp[1])) {
+          mean_temp = (prev_temp + next_temp) / 2
+        } else if (is.na(prev_temp[1]) & !is.na(next_temp[1])) {
+          mean_temp = next_temp
+        } else if (!is.na(prev_temp[1]) & is.na(next_temp[1])) {
+          # if the last packet is corrupted, then impute by last temperature observed
+          mean_temp = prev_temp
+        } else {
+          # safe conduct, if both are NA, then leave it without imputation
+          mean_temp = temp_readings[[i]]
+        }
+       
         temp_readings[[i]] = matrix(mean_temp, nrow = nrow(temp_readings[[i]]),
                                     ncol = 2, byrow = T)
       }
@@ -339,50 +521,38 @@ readParmayMatrix = function(filename, output = c("all", "sf", "dynrange")[1],
   
   # Resample data to defined sampling frequency
   if (any(acc_count > 0) & read_acc) { # accelerometer has been activated
-    required_timepoints = seq(from = acc_timestamps[1], to = acc_timestamps[length(acc_timestamps)], 
-                              by = 1/sf)
+    t_start = acc_timestamps[1]
+    t_end = acc_timestamps[nrow(acc_data)]
+    n_required_timepoints = floor((t_end - t_start) * sf) + 1L
+    required_timepoints = seq(from = acc_timestamps[1], by = 1/sf, 
+                              length.out = n_required_timepoints)
     acc_resampled = resample(raw = acc_data, rawTime = acc_timestamps, 
                              time = required_timepoints, 
                              stop = nrow(acc_data), type = interpolationType)
   }
   if (any(gyro_count > 0) & read_gyro) { # gyroscope has been activated
-    # make sure measurement covers full recording length:
-    if (min(gyro_timestamps) > min(required_timepoints)) {
-      gyro_timestamps = rbind(c(NA, NA), gyro_data)
-    }
-    if (max(gyro_timestamps) < max(required_timepoints)) {
-      gyro_timestamps = c(gyro_timestamps, max(required_timepoints))
-      gyro_data = rbind(gyro_data, c(NA, NA))
-    }
-    gyro_resampled = resample(raw = gyro_data, rawTime = gyro_timestamps, 
-                              time = required_timepoints, 
-                              stop = nrow(gyro_data), type = interpolationType)
+    gyro_resampled = resample_to_grid(
+      data = gyro_data,
+      timestamps = gyro_timestamps,
+      grid = required_timepoints,
+      interpolationType = interpolationType
+    )
   }
-  if (any(temp_count > 0) & read_temp) { # temperature has been activated
-    # make sure measurement covers full recording length:
-    if (min(temp_timestamps) > min(required_timepoints)) {
-      temp_timestamps = rbind(c(NA, NA), temp_data)
-    }
-    if (max(temp_timestamps) < max(required_timepoints)) {
-      temp_timestamps = c(temp_timestamps, max(required_timepoints))
-      temp_data = rbind(temp_data, c(NA, NA))
-    }
-    temp_resampled = resample(raw = temp_data, rawTime = temp_timestamps, 
-                              time = required_timepoints, 
-                              stop = nrow(temp_data), type = interpolationType)
+  if (any(temp_count > 0) & read_temp) {
+    temp_resampled = resample_to_grid(
+      data = temp_data,
+      timestamps = temp_timestamps,
+      grid = required_timepoints,
+      interpolationType = interpolationType
+    )
   }
-  if (any(heart_count > 0) & read_heart) { # heart rate has been activated
-    # make sure measurement covers full recording length:
-    if (min(heart_timestamps) > min(required_timepoints)) {
-      heart_timestamps = rbind(c(NA, NA), heart_data)
-    }
-    if (max(heart_timestamps) < max(required_timepoints)) {
-      heart_timestamps = c(heart_timestamps, max(required_timepoints))
-      heart_data = rbind(heart_data, c(NA, NA))
-    }
-    heart_resampled = resample(raw = heart_data, rawTime = heart_timestamps, 
-                               time = required_timepoints, 
-                               stop = nrow(heart_data), type = interpolationType)
+  if (any(heart_count > 0) & read_heart) {
+    heart_resampled = resample_to_grid(
+      data = heart_data,
+      timestamps = heart_timestamps,
+      grid = required_timepoints,
+      interpolationType = interpolationType
+    )
   }
   
   # build output data -----------
